
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <termios.h>
#include <sys/ioctl.h>
#ifdef SOLARIS
   #include <sys/stropts.h>
#endif

   static void processEnvironment_(char *env, int *add_appletviewer_classpath,
   		char **not_found_message);
   static int openPseudoTerm_(int *master_fd_out, int *slave_fd_out);
   static int getTtyp_(int *master_fd_out, int *slave_fd_out);
   static void initPseudoPipe_(int term_fd);
   static int safeOpen_(const char *path, int oflag);
   static int safeRead_(int fildes, void *buf, size_t nbyte);
   static int safeWrite_(int fildes, void *buf, size_t nbyte);
   static int safeClose_(int fd);
   static int safeExecvp_(const char *file, char *const argv[]);
   static void printMessage_(FILE *stream, char *message, char *line_start);


#ifdef SOLARIS
void setenv(const char *name, const char *value, int overwrite) {
   char *str = malloc(strlen(name) + strlen(value) + 2);
   sprintf(str, "%s=%s", name, value);
   putenv(str);
}
#endif


   #define VERBOSE 128
   #define SHOW_CLASSPATH 256

   #define WTW_ADD_APPLETVIEWER_CLASSPATH 1

   /* Change working directory to argv[3], process environment commands
      in argv[1] and run argv[4] with arguments argv[5] argv[6] ... argv[argc - 1]. */
    int main(int argc, char *argv[]) {
      int master_fd, slave_fd;
      int num_read, total_read;
      pid_t pid, parent_pid, java_pid;
      char buffer[101];
      int flags, verbose, add_appletviewer_classpath;
      char *not_found_message;
   
      if(argc < 5) {
         fprintf(stderr, " ----jGRASP wedge: Wrong number of arguments.\n");
         return 1;
      }
   
      if(sscanf(argv[2], "%d", &flags) != 1) {
         fprintf(stderr, " ----jGRASP wedge internal error: "
            "bad flags argument to wedge.\n");
         return 4;
      }
      verbose = (flags & VERBOSE) != 0;
   
      if(chdir(argv[3]) != 0) {
         fprintf(stderr, " ----jGRASP wedge: Unable to change working directory to "
            "\"%s\".\n", argv[3]);
         return 2;
      }
   
      processEnvironment_(argv[1], &add_appletviewer_classpath,
         	&not_found_message);
   
      /* Open pseudoterminal for i/o. */
      if(!openPseudoTerm_(&master_fd, &slave_fd)) {
         fprintf(stderr, " ----jGRASP wedge: Unable to open pty.\n");
         return 3;
      }
      initPseudoPipe_(master_fd);
   
      /*** Sleep 100ms. Java seems to have trouble if the process
           finished too quickly. */
      usleep(100000);
   
      parent_pid = getpid();
      java_pid = getppid();
   
      /* fork - child forks target while parent does nothing. */
      pid = fork();
      if(pid == 0) /* Child process. */
      {
         int target_pid;
         char *path;
      
         pid = getpid();
         /* fork - parent collects stderr and stdout output from child,
            and feeds stdin to child, while child becomes the executed
            process. */
         target_pid = fork();
      
         if(target_pid != 0) {
            /* Parent process. */
            char eof_char;
            struct termios mode;
            int have_exit_status = 0;
            int exit_status;
            int wait_result;
         
            tcgetattr(master_fd, &mode);
            eof_char = mode.c_cc[VEOF];
         
            /* Convert stdin to non-blocking. */
            fcntl(0, F_SETFL, O_NDELAY);
         
            /* Read until parent is dead and pty is empty. */
            while(1)
            {
               /* Read until pty is empty, an error occurs, or
                  about 1000 characters are read.
                  Maximum output rate assuming 10ms timer precision
                  ~= 1000 chars / .01 sec = 100,000 cps. */
               total_read = 0;
               while(total_read < 1000 && (num_read = safeRead_(master_fd, buffer,
               100)) > 0)
               {
                  /* Feed output to standard error. */
                  safeWrite_(2, buffer, num_read);
                  total_read += num_read;
               }
            
               /* Feed input to pty. */
               while((num_read = safeRead_(0, buffer, 100)) > 0)
                  safeWrite_(master_fd, buffer, num_read);
            
               if(kill(java_pid, 0) != 0 || kill(parent_pid, 0) != 0) {
                  int count = 0;
                  /* jGRASP died or parent died. */
                  /* Soft kill the target process. */
                  kill(target_pid, SIGHUP);
                  /* Wait about 5 seconds to see if the target process will shut down. */
                  while(waitpid(target_pid, &exit_status, WNOHANG) == 0 && count++ < 50)
                     usleep(100000);
                  if(waitpid(target_pid, &exit_status, WNOHANG) == 0)
                     /* Target process is still alive, kill it the hard way. */
                     kill(target_pid, SIGKILL);
                  exit(0);
               }
            
               if((wait_result = waitpid(target_pid, &exit_status, WNOHANG)) != 0) {
                  if(wait_result != -1)
                     have_exit_status = 1;
                  else {
                     fprintf(stderr, "\n ----jGRASP wedge: error waiting for target. Target will be killed.\n"
                        " ----   error no %d \"%s\".\n", errno, strerror(errno));
                     kill(target_pid, SIGKILL);
                  }
                  break;
               }
            
               /* Prevent busy-wait by sleeping 10ms. */
               usleep(10000);
            }
         
            /* Finish reading. */
            while((num_read = safeRead_(master_fd, buffer, 100)) > 0)
            {
               /* Feed output to standard error. */
               safeWrite_(2, buffer, num_read);
            }
         
         
            fflush(stderr);
         
            if(have_exit_status) {
               if(WIFEXITED(exit_status)) {
                  if(verbose || WEXITSTATUS(exit_status) != 0)
                     fprintf(stderr, "\n ----jGRASP wedge: exit code for "
                        "process is %d.\n", WEXITSTATUS(exit_status));
                  else
                     fprintf(stderr, "\n");
               }
               else if(WIFSIGNALED(exit_status)) {
                  fprintf(stderr, "\n ----jGRASP wedge: process died "
                     "on signal %d.\n", WTERMSIG(exit_status));
               }
               else
                  fprintf(stderr, "\n ----jGRASP wedge warning: could "
                     "not get exit code for process.\n");
            }
            else
               fprintf(stderr, "\n ----jGRASP wedge warning: could "
                  "not get exit status for process.\n");
         
         
            /* Close the pty. */
            safeClose_(master_fd);
            safeClose_(slave_fd);
         
            fclose(stderr);
         }
         else {
            /* Child process. */
            if(verbose) {
               fprintf(stderr, " ----jGRASP wedge: pid for process is %d  pids"
                  " for wedge are %d and %d.\n", getpid(), parent_pid, pid);
               if((flags & SHOW_CLASSPATH) != 0) {
                  char *classpath = getenv("CLASSPATH");
                  if(classpath == NULL)
                     classpath = "";
                  fprintf(stderr, " ----JGRASP wedge: CLASSPATH"
                     " is \"%s\".\n", classpath);
               }
            
               fprintf(stderr, " ----jGRASP wedge: working directory is %s\n\n", argv[3]); }
         
            /* Redirect stdin, stderr and stdout to the pty. */
            close(1);    /* Close stdout. */
            dup(slave_fd); /* stdout fd becomes slave_fd_out. */
            close(2);    /* Close stderr. */
            dup(slave_fd); /* stderr fd becomes slave_fd_out. */
            close(0);
            dup(slave_fd);
         
            if(slave_fd > 2)
               close(slave_fd);
         
            /***initPseudoPipe_(1);*/
            if(!isatty(1))
               fprintf(stderr,
                  " ----jGRASP wedge: warning: using a non-pty to collect output.\n"
                  "     Output may not flow smoothly.\n"
                  "     Some output may be lost if process does not flush buffers.\n\n");
         
            if(add_appletviewer_classpath) {
               int i;
               // Don't add the classpath if the command already contains one.
               for(i = 5; i < argc; i++) {
                  if(strcmp(argv[i], "-J-cp") == 0 || strcmp(argv[i], "-J-classpath") == 0) {
                     add_appletviewer_classpath = 0;
                     break; }
               }
            }
         
            if(add_appletviewer_classpath) {
               int i;
               char **args;
               char *cp;
               char *classpath = getenv("CLASSPATH");
               if(classpath == NULL)
                  classpath = "";
               args = malloc((argc - 5 + 4) * sizeof(char *));
               args[0] = argv[4];
               args[1] = "-J-cp";
               cp = malloc((strlen("-J") + strlen(classpath) + 1) * sizeof(char));
               sprintf(cp, "-J");
               strcat(cp, classpath);
               args[2] = cp;
               for(i = 5; i < argc; i++)
                  args[i - 5 + 3] = argv[i];
               args[argc - 5 + 3] = NULL;
               safeExecvp_(argv[4], args);
            }
            else
               safeExecvp_(argv[4], argv + 4);
         
            /* Execution failed. */
            fprintf(stderr, " ----jGRASP wedge: could not execute  %s\n", argv[4]);
            fprintf(stderr, " ----   error number %d.\n", errno);
            switch(errno) {
               case EACCES:
                  fprintf(stderr, " ----   \n ----   Access to target denied.\n"
                     				 " ----   \n ----   Check permissions on target.\n");
                  break;
               case EPERM:
                  fprintf(stderr, " ----   \n ----   Target requires superuser access.\n");
                  break;
               case E2BIG:
                  fprintf(stderr, " ----   \n ----   Argument list is too big.\n");
                  break;
               case ENOEXEC:
                  fprintf(stderr, " ----   \n ----   Target does not appear to be an executable for this system.\n");
                  break;
               case ENAMETOOLONG:
                  fprintf(stderr, " ----   \n ----   Path to target is too long.\n");
                  break;
               case ENOENT:
                  fprintf(stderr, " ----   \n ----   Target does not exist or is not on PATH.\n");
                  if(not_found_message != NULL) {
                     fprintf(stderr, " ----   ");
                     printMessage_(stderr, not_found_message, " ----   ");
                     fprintf(stderr, "\n"); }
                  break;
               case ENOMEM:
                  fprintf(stderr, " ----   \n ----   Not enough memory to execute target.\n");
                  break;
               case ENOTDIR:
                  fprintf(stderr, " ----   \n ----   Path to target includes a non-directory.\n");
                  break;
               case ELOOP:
                  fprintf(stderr, " ----   \n ----   Too many symbolic links in path to target.\n");
                  break;
               case EIO:
                  fprintf(stderr, " ----   \n ----   I/O error running target.\n");
                  break;
               case ENFILE:
                  fprintf(stderr, " ----   \n ----   Too many files open on system.\n");
                  break;
            }    
         		
            if(argv[4][0] != '/') {
               path = getenv("PATH");
               if(path == NULL)
                  path = "";
               fprintf(stderr, " ----   \n ----   PATH is \"%s\".\n", path); }
         
            fflush(stderr);
            fflush(stdout);
            _exit(0);
         
         }
      }
      
      else /* Parent process. */
      {
         int wait_result, exit_status;
      
         /* Do nothing, just wait for the child to die. */
         while(1) {
            if((wait_result = waitpid(pid, &exit_status, WNOHANG)) != 0) {
                  //fprintf(stderr, "\n***wait result %d %d\n", wait_result, errno);
               if(wait_result == -1) {
                  fprintf(stderr, "\n ----jGRASP wedge: error waiting for second half of wedge.\n"
                     " ----   Target may remain running.\n"
                     " ----   error no %d \"%s\".\n", errno, strerror(errno));
               }               
               break;
            }
            usleep(10000);
         }
         return 0;
      }
   
      return 0;
   }



    static void processEnvironment_(char *env, int *add_appletviewer_classpath,
    		char **not_found_message) {
      int on_val;
      int env_len = strlen(env);
      unsigned int pos = 0;
      char type;
      int val_start;
      unsigned int buffer_size = 1000;
      char *buffer, *current_val;
      int i;
      char ch, ch2;
      unsigned int end_pos;
   
      *not_found_message = NULL;
      *add_appletviewer_classpath = 0;
      if(env_len <= 2)
         return;
   
      /* Process enviroment. */
      on_val = 0;
      buffer = (char *)malloc(buffer_size);
      for(i = 0; i < env_len; i++) {
         if(pos + 1 >= buffer_size) {
            buffer_size *= 2;
            buffer = (char *)realloc(buffer, buffer_size); }
         ch = env[i];
         if(env[i] == '\\') {
            ch2 = env[++i];
            if(on_val && (ch2 == 'n' || ch2 == 'e')) {
               on_val = 0;
               buffer[pos] = '\0';
               end_pos = pos;
               pos = 0;
               if(type == 'i') {
                  if(strcmp(buffer, "NOT_FOUND_MESSAGE") == 0) {
                     *not_found_message = buffer + val_start; }
                  else if(strcmp(buffer, "ADD_APPLETVIEWER_CLASSPATH") == 0) {
                     *add_appletviewer_classpath = 1; }
               }
               else {
                  if(type != 'r') {
                     buffer[val_start - 1] = '\0';
                     current_val = getenv(buffer);
                     if(current_val != NULL) {
                        int size = strlen(current_val);
                        if(buffer_size < size + end_pos + 1) {
                           buffer_size = size + end_pos + 1;
                           buffer = (char *)realloc(buffer, buffer_size); }
                        if(type == 'a') {
                           memmove(buffer + val_start + size,
                              buffer + val_start, end_pos - val_start);
                           memcpy(buffer + val_start, current_val, size);
                           buffer[end_pos + size] = '\0';
                        }
                        else
                           sprintf(buffer + end_pos, "%s", current_val);
                     }
                  }
                  setenv(buffer, buffer + val_start, 1); }
               continue;
            }
            else if(!on_val && (ch2 == 'r' || ch2 == 'p' || ch2 == 'a' || ch2 == 'i')) {
               on_val = 1;
               buffer[pos++] = '\0';/*'=';*/
               val_start = pos;
               type = ch2;
               continue;
            }
            else if(ch2 == 'd')
               ch = '\"';
            else if(ch2 == 'q')
               ch = '\'';
            else
               ch = ch2;
         }
         buffer[pos++] = ch;
      }
      free(buffer);
   }



   /* Open a terminal for unidirectional use. */
    static int openPseudoTerm_(int *master_fd_out, int *slave_fd_out)
   {
      int master_fd, slave_fd;
      char *slave_name;
   
      #ifndef SOLARIS
      if(openpty(&master_fd, &slave_fd, NULL, NULL, NULL) == 0) {
         fcntl(master_fd, F_SETFL, O_NDELAY | O_NONBLOCK);
         *master_fd_out = master_fd;
         *slave_fd_out = slave_fd;
         return 1;
      }
      #endif
      master_fd = -1;
   
   /***LAB Try moving  O_NDELAY flags into opens and get rid of fcntl below. */
      #if defined(SOLARIS) || defined(DEC_ALPHA)
      if((master_fd = safeOpen_("/dev/ptmx", O_RDWR)) > 0) {
         grantpt(master_fd);
         unlockpt(master_fd);
         slave_name = ptsname(master_fd);
      }
      #elif defined(AIX)
      if((master_fd = safeOpen_("/dev/ptc", O_RDWR)) > 0)
         slave_name = ttyname(master_fd);
      #elif defined(IRIX4) || defined(IRIX6)
      slave_name = _getpty(&master_fd, O_RDWR | O_NOCTTY | O_NDELAY, 0600, 0);
      if(slave_name == NULL)
         master_fd = -1;
      #endif
   
      if(master_fd >= 0) {
         if((slave_fd = safeOpen_(slave_name, O_RDWR)) > 0) {
            #ifdef SOLARIS
            ioctl(slave_fd, I_PUSH, "ptem");
            #endif
         
            fcntl(master_fd, F_SETFL, O_NDELAY | O_NONBLOCK);
            *master_fd_out = master_fd;
            *slave_fd_out = slave_fd;
            return 1;
         }
         /* Failed to open slave. */
         close(master_fd);
      }
   
   
      /* Try a ttyp. */
      if(getTtyp_(master_fd_out, slave_fd_out))
         return 2;
   
   
      return 0;
   }




   /* Open a tty style terminal.
      Returns non-zero on success. */
   static char master_name[] = "/dev/ptyxx";
   static char slave_name[] = "/dev/ttyxx";
   static char index1_chars[] = "pqrstuvwxyz";
   static char index2_chars[] = "0123456789abcdef";
    static int getTtyp_(int *master_fd_out, int *slave_fd_out)
   {
      int master_fd, slave_fd;
      char *c3, *c4;
   
      for (c3 = index1_chars; *c3 != 0; c3++) {
         for (c4 = index2_chars; *c4 != 0; c4++) {
            master_name[8] = slave_name[8] = *c3;
            master_name[9] = slave_name[9] = *c4;
            /* Try to open the master. */
            if((master_fd = safeOpen_(master_name, O_RDWR)) >= 0) 
            {
               /* If slave can be opened, we are done. */
               if((slave_fd = safeOpen_(slave_name, O_RDWR)) >= 0)  {
                  #ifdef SOLARIS
                  ioctl(slave_fd, I_PUSH, "ptem");
                  #endif
               
                  fcntl(master_fd, F_SETFL, O_NDELAY | O_NONBLOCK);
                  *master_fd_out = master_fd;
                  *slave_fd_out = slave_fd;
                  return 1;
               }
            
               /* Slave inaccessible. Close master and keep going. */
               close(master_fd);
            }
         }
      }
   
      return 0;
   }




   /* Initialize a terminal to be used for output collection
      only. */
    static void initPseudoPipe_(int term_fd)
   {
      struct termios mode;
      struct winsize win_size;
   
      if(!isatty(term_fd)) /* No initialization necessary for pipe. */
         return;
   
      #ifdef TIOCGETA
      ioctl(term_fd, TIOCGETA, (char *)&mode);
      #elif defined(TCGETS)
      ioctl(term_fd, TCGETS, (char *)&mode);
      #else
      tcgetattr(term_fd, &mode);
      #endif        
   
      mode.c_lflag = 0; /* No signal processing. */      
      mode.c_oflag = 0; /* No post-processing of output. */
      mode.c_cflag = B9600 | CS8 | CREAD; /* 8 bit chars. */
   
      #ifdef TIOCSETA
      ioctl(term_fd, TIOCSETA, (char *)&mode);
      #elif defined(TCSETS)
      ioctl(term_fd, TCSETS, (char *)&mode);
      #else
   
      tcsetattr(term_fd, TCSANOW, &mode);
      #endif
   
      /* We could use the width of the message window, but 80
         characters should be sufficient. */
      win_size.ws_row = (unsigned short)80;
      win_size.ws_col = (unsigned short)10;
      ioctl(term_fd, TIOCSWINSZ, (char *)&win_size);
   }


    static int safeOpen_(const char *path, int oflag)
   {
      int retval;
      while((retval = open(path, oflag)) == -1 && errno == EINTR);
      return retval;
   }


    static int safeRead_(int fildes, void *buf, size_t nbyte)
   {
      int retval;
      while((retval = read(fildes, buf, nbyte)) == -1 && errno == EINTR);
      return retval;
   }



    static int safeWrite_(int fildes, void *buf, size_t nbyte)
   {
      int retval;
      while((retval = write(fildes, buf, nbyte)) == -1 && errno == EINTR);
      return retval;
   }



    static int safeClose_(int fd)
   {
      int retval;
      while((retval = close(fd)) == -1 && errno == EINTR);
      return retval;
   }




#ifdef IRIX4
   /* argv parameter is incorrectly qualified on IRIX 4. */
   static int safeExecvp_(const char *file, char *const argv[])
   {
      while(execvp(file, (const char * [])argv) == -1 && errno == EINTR);
      return 1;
   }
#else
    static int safeExecvp_(const char *file, char *const argv[])
   {
      while(execvp(file, argv) == -1 && errno == EINTR);
      return 1;
   }
#endif


    static void printMessage_(FILE *stream, char *message, char *line_start) {
      int i;
      int message_len;
    
      message_len = strlen(message);
      for(i = 0; i < message_len; i++) {
         if(message[i] == '\\') {
            if(i + 1 < message_len) {
               i++;
               if(message[i] == 'n')
                  fprintf(stream, "\n%s", line_start);
               else
                  fprintf(stream, "%c", message[i]);
            }
            else
               fprintf(stream, "\\");
         }
         else
            fprintf(stream, "%c", message[i]);
      }
   }
